# -*- encoding: utf-8 -*-
import os
from pathlib import Path
from ddcUtils import constants


def get_current_path() -> Path:
    path = os.path.abspath(os.getcwd())
    return Path(os.path.normpath(path)) if path else None


def get_pictures_path() -> Path:
    if constants.OS_NAME == "Windows":
        import winreg
        sub_key = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders"
        pictures_guid = "My Pictures"
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, sub_key) as key:
            pictures_path = winreg.QueryValueEx(key, pictures_guid)[0]
        return Path(pictures_path)
    else:
        pictures_path = os.path.join(os.getenv("HOME"), "Pictures")
        return Path(pictures_path)


def get_downloads_path() -> Path:
    if constants.OS_NAME == "Windows":
        import winreg
        sub_key = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders"
        downloads_guid = "{374DE290-123F-4565-9164-39C4925E467B}"
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, sub_key) as key:
            downloads_path = winreg.QueryValueEx(key, downloads_guid)[0]
        return Path(downloads_path)
    else:
        downloads_path = os.path.join(os.getenv("HOME"), "Downloads")
        return Path(downloads_path)



import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

__version_info__ = ("1", "0", "0")
__version__ = ".".join(__version_info__)
__author__ = "ddc"
__email__ = "ddc@users.noreply.github.com"
__req_python_version__ = (3, 11, 0)



# -*- encoding: utf-8 -*-
import sys


def get_exception(e) -> str:
    module = e.__class__.__module__
    if module is None or module == str.__class__.__module__:
        module_and_exception = f"[{e.__class__.__name__}]:[{e}]"
    else:
        module_and_exception = f"[{module}.{e.__class__.__name__}]:[{e}]"
    return module_and_exception.replace("\r\n", " ").replace("\n", " ")


class CustomBaseException(Exception):
    def __init__(self, msg):
        sys.stderr.write(get_exception(msg))


class DBAddException(CustomBaseException):
    pass


class DBExecuteException(CustomBaseException):
    pass


class DBFetchAllException(CustomBaseException):
    pass


class DBFetchValueException(CustomBaseException):
    pass



# -*- encoding: utf-8 -*-
from pathlib import Path
import platform
import sys
from . import (
    __version__,
    __req_python_version__,
)


VERSION = __version__
PYTHON_OK = sys.version_info >= __req_python_version__
MIN_PYTHON_VERSION = ".".join(str(x) for x in __req_python_version__)
OS_NAME = platform.system()
DATE_TIME_FORMATTER_STR = "%a %b %m %Y %X"
DATE_FORMATTER = "%Y-%m-%d"
TIME_FORMATTER = "%H:%M:%S.%f"
BASE_DIR = Path(__file__).resolve().parent.parent



# -*- encoding: utf-8 -*-
import configparser
import errno
import gzip
import os
from pathlib import Path
import shutil
import struct
import subprocess
import sys
import zipfile
import requests
from ddcUtils import constants
from ddcUtils.exceptions import get_exception


def open_file(file_path: str) -> int:
    if not os.path.isfile(file_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
    try:
        match constants.OS_NAME:
            case "Windows":
                os.startfile(file_path)
                return_code = 0
            case "Darwin":
                return_code = subprocess.call(("open", file_path))
            case _:
                return_code = subprocess.call(("xdg-open", file_path))
        return return_code
    except Exception as e:
        sys.stderr.write(get_exception(e))
        return 1


def list_files(directory: str, file_extension: str = None) -> list:
    files_list = None
    if os.path.isdir(directory):
        if not file_extension:
            files_list = [Path(os.path.join(directory, f)) for f in os.listdir(directory)]
        else:
            files_list = [Path(os.path.join(directory, f)) for f in os.listdir(directory) if
                          f.lower().endswith(file_extension.lower())]
        files_list.sort(key=os.path.getctime)
    return files_list


def gzip_file(file_path: str) -> Path | None:
    file_name = os.path.basename(file_path)
    gz_out_file_path = os.path.join(os.path.dirname(file_path), f"{file_name}.gz")

    try:
        with open(file_path, "rb") as fin:
            with gzip.open(gz_out_file_path, "wb") as fout:
                fout.writelines(fin)
        return Path(gz_out_file_path)
    except Exception as e:
        sys.stderr.write(get_exception(e))
        if os.path.isfile(gz_out_file_path):
            os.remove(gz_out_file_path)
        return None


def unzip_file(file_path: str, out_path: str = None) -> zipfile.ZipFile | None:
    try:
        out_path = out_path or os.path.dirname(file_path)
        zipfile_path = file_path
        zipf = zipfile.ZipFile(zipfile_path)
        zipf.extractall(out_path)
        zipf.close()
        return zipf
    except Exception as e:
        sys.stderr.write(get_exception(e))
        return None


def get_exe_binary_type(file_path: str) -> str | None:
    with open(file_path, "rb") as f:
        s = f.read(2)
        if s != b"MZ":
            return "Not an EXE file"
        f.seek(60)
        s = f.read(4)
        header_offset = struct.unpack("<L", s)[0]
        f.seek(header_offset + 4)
        s = f.read(2)
        machine = struct.unpack("<H", s)[0]
        match machine:
            case 332:
                sys.stdout.write("IA32 (32-bit x86)")
                binary_type = "IA32"
            case 512:
                sys.stdout.write("IA64 (Itanium)")
                binary_type = "IA64"
            case 34404:
                sys.stdout.write("AMD64 (64-bit x86)")
                binary_type = "AMD64"
            case 452:
                sys.stdout.write("ARM eabi (32-bit)")
                binary_type = "ARM-32bits"
            case 43620:
                sys.stdout.write("AArch64 (ARM-64, 64-bit)")
                binary_type = "ARM-64bits"
            case _:
                sys.stdout.write(f"Unknown architecture {machine}")
                binary_type = None
    return binary_type


def _get_default_parser() -> configparser.ConfigParser:
    parser = configparser.ConfigParser(delimiters="=", allow_no_value=True)
    parser.optionxform = str  # this will not change all values to lowercase
    parser._interpolation = configparser.ExtendedInterpolation()
    return parser


def _get_parser_value(parser, section: str, config_name: str) -> str | int | None:
    try:
        value = parser.get(section, config_name).replace("\"", "")
        lst_value = list(value.split(","))
        if len(lst_value) > 1:
            values = []
            for each in lst_value:
                values.append(int(each.strip()) if each.strip().isnumeric() else each.strip())
            value = values
        elif value is not None and type(value) is str:
            if len(value) == 0:
                value = None
            elif value.isnumeric():
                value = int(value)
            elif "," in value:
                value = sorted([x.strip() for x in value.split(",")])
        else:
            value = None
    except Exception as e:
        sys.stderr.write(get_exception(e))
        value = None
    return value


def get_all_file_values(file_path: str, mixed_values: bool = False) -> dict:
    if not os.path.isfile(file_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
    final_data = {}
    parser = _get_default_parser()
    try:
        parser.read(file_path)
        for section in parser.sections():
            section_name = section.lower().replace(" ", "_")
            if not mixed_values:
                final_data[section_name] = {}
            for name in parser.options(section):
                config_name = name.lower().replace(" ", "_")
                value = _get_parser_value(parser, section, config_name)
                if not mixed_values:
                    final_data[section_name][config_name] = value
                else:
                    if config_name in final_data:
                        config_name = f"{section_name}_{config_name}"
                    final_data[config_name] = value
        return final_data if len(final_data) > 0 else None
    except Exception as e:
        sys.stderr.write(get_exception(e))


def get_all_file_section_values(file_path: str, section: str) -> dict:
    if not os.path.isfile(file_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
    final_data = {}
    parser = _get_default_parser()
    try:
        parser.read(file_path)
        for name in parser.options(section):
            config_name = name.lower().replace(" ", "_")
            value = _get_parser_value(parser, section, config_name)
            final_data[config_name] = value
        return final_data if len(final_data) > 0 else None
    except Exception as e:
        sys.stderr.write(get_exception(e))


def get_file_value(file_path: str, section: str, config_name: str) -> str | int | None:
    if not os.path.isfile(file_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
    parser = _get_default_parser()
    parser.read(file_path)
    value = _get_parser_value(parser, section, config_name)
    return value


def set_file_value(file_path: str, section: str, config_name: str, value) -> bool:
    if not os.path.isfile(file_path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), file_path)
    parser = _get_default_parser()
    parser.read(file_path)
    new_value = value
    if isinstance(value, str):
        new_value = f'"{value}"'
    parser.set(section, config_name, new_value)
    try:
        with open(file_path, "w") as configfile:
            parser.write(configfile, space_around_delimiters=False)
        return True
    except configparser.DuplicateOptionError:
        return False


def copydir(src, dst, symlinks=False, ignore=None) -> bool:
    try:
        for item in os.listdir(src):
            s = os.path.join(src, item)
            d = os.path.join(dst, item)
            if os.path.isdir(s):
                shutil.copytree(s, d, symlinks, ignore)
            else:
                shutil.copy2(s, d)
        return True
    except IOError as e:
        sys.stderr.write(get_exception(e))
    return False


def download_file(remote_file_url, local_file_path) -> bool:
    try:
        req = requests.get(remote_file_url)
        if req.status_code == 200:
            with open(local_file_path, "wb") as outfile:
                outfile.write(req.content)
            return True
    except requests.HTTPError as e:
        sys.stderr.write(get_exception(e))
    return False



# -*- encoding: utf-8 -*-
import gzip
import logging.handlers
import os
import sys


class Log:
    def __init__(
        self,
        app_name: str = "app",
        dir_logs: str = "logs",
        days_to_keep: int = 7,
        level: str = "info"
    ):
        self.app_name = app_name
        self.dir = dir_logs
        self.days_to_keep = days_to_keep
        self.level = _get_level(level)

    def setup_logging(self):
        try:
            os.makedirs(self.dir, exist_ok=True) if not os.path.isdir(self.dir) else None
        except Exception as e:
            sys.stderr.write(f"[ERROR]:[Unable to create logs dir]:{str(e)}: {self.dir}\n")
            sys.exit(1)

        log_file_path = os.path.join(self.dir, f"{self.app_name}.log")

        try:
            open(log_file_path, "a+").close()
        except IOError as e:
            sys.stderr.write(f"[ERROR]:[Unable to open log file for writing]:{str(e)}: {log_file_path}\n")
            sys.exit(1)

        if self.level == logging.DEBUG:
            formatt = "[%(asctime)s.%(msecs)03d]:[%(levelname)s]:[%(filename)s:%(funcName)s:%(lineno)d]:%(message)s"
        else:
            formatt = "[%(asctime)s.%(msecs)03d]:[%(levelname)s]:%(message)s"

        formatter = logging.Formatter(formatt, datefmt="%Y-%m-%dT%H:%M:%S")

        logger = logging.getLogger()
        logger.setLevel(self.level)

        file_hdlr = logging.handlers.TimedRotatingFileHandler(filename=log_file_path,
                                                              encoding="UTF-8",
                                                              when="midnight",
                                                              backupCount=self.days_to_keep)

        file_hdlr.setFormatter(formatter)
        file_hdlr.suffix = "%Y%m%d"
        file_hdlr.rotator = GZipRotator(self.dir, self.days_to_keep)
        file_hdlr.setLevel(self.level)
        logger.addHandler(file_hdlr)

        stream_hdlr = logging.StreamHandler()
        stream_hdlr.setFormatter(formatter)
        stream_hdlr.setLevel(self.level)
        logger.addHandler(stream_hdlr)

        return logger


class GZipRotator:
    def __init__(self, dir_logs, days_to_keep):
        self.dir = dir_logs
        self.days_to_keep = days_to_keep

    def __call__(self, source, dest):
        RemoveOldLogs(self.dir, self.days_to_keep)
        if os.path.isfile(source) and os.stat(source).st_size > 0:
            try:
                sfname, sext = os.path.splitext(source)
                _, dext = os.path.splitext(dest)
                renamed_dst = f"{sfname}_{dext.replace('.', '')}{sext}.gz"
                with open(source, "rb") as fin:
                    with gzip.open(renamed_dst, "wb") as fout:
                        fout.writelines(fin)
                os.remove(source)
            except Exception as e:
                sys.stderr.write(f"[ERROR]:[Unable to zip log file]:{str(e)}: {source}\n")


class RemoveOldLogs:
    def __init__(self, logs_dir, days_to_keep):
        files_list = [f for f in os.listdir(logs_dir)
                      if os.path.isfile(f"{logs_dir}/{f}") and os.path.splitext(f)[1] == ".gz"]
        for file in files_list:
            file_path = f"{logs_dir}/{file}"
            if self._is_file_older_than_x_days(file_path, days_to_keep):
                try:
                    os.remove(file_path)
                except Exception as e:
                    sys.stderr.write(f"[ERROR]:[Unable to remove old logs]:{str(e)}: {file_path}\n")

    @staticmethod
    def _is_file_older_than_x_days(file_path, days_to_keep):
        from datetime import datetime, timedelta
        file_time = datetime.fromtimestamp(os.path.getctime(file_path))
        if int(days_to_keep) == 1:
            cutoff_time = datetime.today()
        else:
            cutoff_time = datetime.today() - timedelta(days=int(days_to_keep))
        file_time = file_time.replace(hour=0, minute=0, second=0, microsecond=0)
        cutoff_time = cutoff_time.replace(hour=0, minute=0, second=0, microsecond=0)
        if file_time < cutoff_time:
            return True
        return False


def _get_level(level: str):
    if not isinstance(level, str):
        return logging.INFO
    match level.lower():
        case "debug":
            return logging.DEBUG
        case "warning":
            return logging.WARNING
        case "error":
            return logging.ERROR
        case "critical":
            return logging.CRITICAL
        case _:
            return logging.INFO



# -*- encoding: utf-8 -*-
from ..exceptions import (
    DBAddException,
    DBExecuteException,
    DBFetchAllException,
    DBFetchValueException
)


class DBUtils:
    def __init__(self, session):
        self.session = session

    def add(self, stmt):
        try:
            self.session.add(stmt)
        except Exception as e:
            self.session.rollback()
            raise DBAddException(e)
        else:
            self.session.commit()

    def execute(self, stmt):
        try:
            self.session.execute(stmt)
        except Exception as e:
            self.session.rollback()
            raise DBExecuteException(e)
        else:
            self.session.commit()

    def fetchall(self, stmt):
        cursor = None
        try:
            cursor = self.session.execute(stmt)
        except Exception as e:
            self.session.rollback()
            raise DBFetchAllException(e)
        else:
            self.session.commit()
            return cursor.mappings().all()
        finally:
            cursor.close() if cursor is not None else None

    def fetchone(self, stmt):
        cursor = None
        try:
            cursor = self.session.execute(stmt)
        except Exception as e:
            self.session.rollback()
            raise DBFetchAllException(e)
        else:
            self.session.commit()
            return cursor.mappings().first()
        finally:
            cursor.close() if cursor is not None else None

    def fetch_value(self, stmt):
        cursor = None
        try:
            cursor = self.session.execute(stmt)
        except Exception as e:
            self.session.rollback()
            raise DBFetchValueException(e)
        else:
            self.session.commit()
            return cursor.first()[0]
        finally:
            cursor.close() if cursor is not None else None



# -*- encoding: utf-8 -*-
import sys
import sqlalchemy as sa
from sqlalchemy.engine import create_engine
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine, async_sessionmaker
from sqlalchemy.orm import Session
from ..exceptions import get_exception


class DBSqlite:
    def __init__(self, db_file_path, batch_size=100, echo=False):
        self.file = db_file_path
        self.batch_size = batch_size
        self.echo = echo

    def get_sqlite_engine(self):
        try:
            engine = create_engine(f"sqlite:///{self.file}", future=True, echo=self.echo).\
                execution_options(stream_results=self.echo,
                                  isolation_level="AUTOCOMMIT")

            @sa.event.listens_for(engine, "before_cursor_execute")
            def receive_before_cursor_execute(conn,
                                              cursor,
                                              statement,
                                              params,
                                              context,
                                              executemany):
                cursor.arraysize = self.batch_size
            return engine
        except Exception as e:
            sys.stderr.write(f"Unable to Create Database Connection: {get_exception(e)}")
            return None

    @staticmethod
    def get_db_session(engine):
        session = Session(bind=engine)
        return session


class DBPostgres:
    def __init__(self, **kwargs):
        self.username = kwargs["username"]
        self.password = kwargs["password"]
        self.host = kwargs["host"]
        self.port = kwargs["port"]
        self.db = kwargs["database"]

    def _set_engine(self):
        return create_async_engine(
            self.get_uri(),
            echo=False,
            future=True
        )

    def get_uri(self):
        credentials = {
            "drivername": "postgresql+asyncpg",
            "username": self.username,
            "password": self.password,
            "host": self.host,
            "port": self.port,
            "database": self.db
        }
        return sa.engine.URL.create(**credentials)

    def get_db_engine(self):
        return self._set_engine()

    @staticmethod
    def get_db_session(async_engine):
        async_session = async_sessionmaker(
            bind=async_engine,
            autoflush=True,
            expire_on_commit=False,
            future=True,
            class_=AsyncSession
        )
        return async_session()



# -*- encoding: utf-8 -*-
from datetime import datetime, timezone
import json
import os
from pathlib import Path
from ddcUtils import constants


class Object:
    def __init__(self):
        self._created = datetime.now().isoformat()

    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

    def to_dict(self):
        json_string = json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)
        json_dict = json.loads(json_string)
        return json_dict


def clear_screen() -> None:
    if constants.OS_NAME == "Windows":
        os.system("cls")
    else:
        os.system("clear")


def user_choice() -> input:
    try:
        return input(">>> ").lower().strip()
    except SyntaxError:
        pass


def get_active_branch_name(default_master_branch_name: str = "master") -> str:
    head_dir = Path(os.path.join(constants.BASE_DIR, ".git", "HEAD"))
    try:
        with head_dir.open("r") as f:
            content = f.read().splitlines()
        for line in content:
            if line[0:4] == "ref:":
                return line.partition("refs/heads/")[2]
    except FileNotFoundError:
        return default_master_branch_name


def get_current_date_time() -> datetime:
    return datetime.now(timezone.utc)


def get_current_date_time_str_long() -> str:
    return convert_datetime_to_str_long(get_current_date_time())


def convert_datetime_to_str_long(date: datetime) -> str:
    return date.strftime(constants.DATE_TIME_FORMATTER_STR)


def convert_datetime_to_str_short(date: datetime) -> str:
    return date.strftime(f"{constants.DATE_FORMATTER} {constants.TIME_FORMATTER}")


def convert_str_to_datetime_short(date_str: str) -> datetime:
    return datetime.strptime(date_str, f"{constants.DATE_FORMATTER} {constants.TIME_FORMATTER}")
